@@ -9,12 +9,12 @@ use Scalar::Util qw ( blessed );
 use Data::Dumper;
 use Attribute::Handlers;
 
-our $VERSION    = '1.01';
+our $VERSION    = '1.02';
 my  $SIGNATURES = {};
 
 sub UNIVERSAL::with : ATTR(CODE,INIT) {
   my ($package, $symbol, $referent, $attr, $data) = @_;
- 
+
   my $large   = *{$symbol}{NAME};
   my $subname = substr($large, rindex($large, ':') + 1);
 
@@ -24,9 +24,9 @@ sub UNIVERSAL::with : ATTR(CODE,INIT) {
   if (!ref($data)) {
     $data = [ $data ];
   }
-  
+
   ## save this for later use
-  $SIGNATURES->{$package}->{$subname} = $data;
+  $SIGNATURES->{$package}->{with}->{$subname} = $data;
 
   my $attributes = { map { ($_, 1) } attributes::get( $referent ) };
 
@@ -76,7 +76,7 @@ sub UNIVERSAL::with : ATTR(CODE,INIT) {
     if ($attributes->{method}) { $m++; }
 
     print "Out of band:\n\tCount\tMatched\n\t$count\t$m\n" if $::AS_DEBUG;
-    
+
     if ($m != $count) {
       croak("call to $subname does not match signature");
     } else {
@@ -85,6 +85,83 @@ sub UNIVERSAL::with : ATTR(CODE,INIT) {
   };
 }
 
+sub UNIVERSAL::returns : ATTR(CODE,INIT) {
+  my ($package, $symbol, $referent, $attr, $data) = @_;
+
+  my $large   = *{$symbol}{NAME};
+  my $subname = substr($large, rindex($large, ':') + 1);
+
+  no warnings qw( redefine );
+
+  ## make sure we have an array ref, so its easier
+  if (!ref($data)) {
+    $data = [ $data ];
+  }
+
+  ## save this for later use
+  $SIGNATURES->{$package}->{returns}->{$subname} = $data;
+
+  my $attributes = { map { ($_, 1) } attributes::get( $referent ) };
+
+  if ($attributes->{method}) {
+    print "Signature on sub $subname is for a method\n" if $::AS_DEBUG;
+    unshift @$data, $package;  ## put a placeholder in the front
+  }
+
+  *{$symbol} = sub {
+
+    my @return = $referent->( @_ );
+
+    my $i = 0;
+    my $count = scalar(@return);
+
+    if ($count != scalar(@$data)) {
+      if ($attributes->{method}) {
+	croak("invalid number of arguments returned from method $subname");
+      } else {
+	croak("invalid number of arguments returned from subroutine $subname");
+      }
+    }
+
+    my $m = 0;
+    print "ReturnComparisons\n" if $::AS_DEBUG;
+    print "\tSignature\tValue\n" if $::AS_DEBUG;
+    while($i <= $count) {
+      print "\t$data->[$i]\t\t$return[$i]\n" if $::AS_DEBUG;
+      last unless $data->[$i];
+      if (lc($data->[$i]) eq $data->[$i]) {
+	## here we are checking for little types
+	my $type = $data->[$i];
+	if (Attribute::Signature->can( $type )) {
+	  if (Attribute::Signature->$type( $return[$i] )) {
+	    $m++;
+	  }
+	}
+      } elsif ($data->[$i] eq 'REF' && ref($return[$i])) {
+	$m++;
+#      } elsif ((blessed($return[$i]) || string($return[$i])) && $return[$i]->isa( $data->[$i]) ) {
+      } elsif (blessed($return[$i]) && $return[$i]->isa( $data->[$i]) ) {
+	$m++;
+      } elsif (!blessed($return[$i]) && ref($return[$i]) eq $data->[$i]) {
+	$m++;
+      } else {
+	# no match
+      }
+      $i++;
+    }
+
+    if ($attributes->{method}) { $m++; }
+
+    print "ReturnOut of band:\n\tCount\tMatched\n\t$count\t$m\n" if $::AS_DEBUG;
+
+    if ($m != $count) {
+      croak("Arguments returned from $subname do not match signature $m != $count");
+    } else {
+      $referent->( @_ );
+    }
+  };
+}
+
 sub getSignature  {
   my $class = shift;
   my $fqsn  = shift;
@@ -93,28 +170,31 @@ sub getSignature  {
   my $subname = substr($fqsn, rindex($fqsn, ':') + 1);
   my $package = substr($fqsn, 0, rindex($fqsn, '::'));
 
-  return $SIGNATURES->{$package}->{$subname};
+  if (wantarray) {
+    return($SIGNATURES->{$package}->{with}->{$subname}, $SIGNATURES->{$package}->{returns}->{$subname});
+  } else {
+    return $SIGNATURES->{$package}->{with}->{$subname};
+  }
 }
 
 sub string {
-  return 1 unless ref($_[0]);
-  return 0;
+  my $class = shift;
+  return not ref $_[0];
 }
 
 sub number {
-  return 1 if (float($_[0]) || int($_[0]));
+  my $class = shift;
+  return $class->float($_[0]) || $class->integer($_[0]);
 }
 
 sub float {
   my $class = shift;
-  return 1 if $_[0] =~ /^\d*\.\d*$/;
-  return 0;
+  return $_[0] =~ /^\d*\.\d*$/;
 }
 
-sub int {
+sub integer {
   my $class = shift;
-  return 1 if $_[0] =~ /^\d+$/;
-  return 0;
+  return $_[0] =~ /^\d+$/;
 }
 
 1;
@@ -130,22 +210,27 @@ Attribute::Signature - allows you to define a call signature for subroutines
   package Some::Package;
   use Attribute::Signature;
 
-  sub somesub : with(float, string, Some::Other::Class) {
+  sub somesub : with(float, string, Some::Other::Class) returns(float) {
     # .. do something ..
   }
 
   package main;
-  my $array = Attribute::Signature->getSignature( 'Some::Package::somesub' );
+  my $array = Attribute::Signature->getSignature('Some::Package::somesub');
 
 =head1 DESCRIPTION
 
-This module allows you to declare a calling signature for a method.  As yet it
-does not provide multimethod type functionality, but it does prevent you from
-writing lots of annoying code to check argument types inside your subroutine.
-C<Attribute::Signature> takes two forms, the first is attributes on standard
-subroutines, in which it examines every parameter passed to the subroutine.  However,
-if the subroutine is marked with the method attribute, then Attribute::Signature
-will not examine the first argument, which can be either the class or the instance.
+This module allows you to declare calling and returning signatures for
+a method.  As yet it does not provide multimethod type functionality,
+but it does prevent you from writing lots of annoying code to check
+argument types inside your subroutine.  C<Attribute::Signature> takes
+two forms, the first is attributes on standard subroutines, in which
+it examines every parameter passed to the subroutine.  However, if the
+subroutine is marked with the method attribute, then
+Attribute::Signature will not examine the first argument, which can be
+either the class or the instance.
+
+C<Attribute::Signature> can also  check  the values that  are returned
+from the method / subroutine, by use of the C<returns> attribute.
 
 C<Attribute::Signature> checks for the following types:
 
@@ -181,7 +266,7 @@ namespace.  Attribute::Signature comes with the following type tests:
 
 =item float
 
-=item int
+=item integer
 
 =item string
 
@@ -189,19 +274,23 @@ namespace.  Attribute::Signature comes with the following type tests:
 
 =back
 
-You can define more tests by declaring subs in the Attribute::Signature namespace.
+Note that the float type mistakenly decides that 10.0 is not a float
+as Perl optimises it to be 10. You can define more tests by declaring
+subs in the Attribute::Signature namespace.
 
 =head1 OTHER FUNCTIONS
 
 =item getSignature( string )
 
-C<Attribute::Signature> also allows you to call the getSignature method.  The string
-should be the complete namespace and subroutine.
-This returns the attribute signature for the function as an array reference.
+C<Attribute::Signature> also allows you to call the getSignature
+method.  The string should be the complete namespace and subroutine.
+This returns the attribute signature and returned values signature for
+the function as two array references.
 
 =head1 AUTHOR
 
 James A. Duncan <jduncan@fotango.com>
+Leon Brocard <leon@fotango.com>
 
 =head1 SEE ALSO
 
